/*
 * cache.c: MMU and cache operations for the LX4xxx/5xxx
 *
 *
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <asm/mipsregs.h>
#include <asm/page.h>
#include <asm/pgtable.h>
#include <asm/system.h>
#include <asm/mmu_context.h>
#include <asm/isadep.h>
#include <asm/io.h>
#include <asm/wbflush.h>
#include <asm/cpu.h>
#include <asm/asm.h>
#include <asm/cacheops.h>

/**
 * This structure contains processor related information.
 * The tlbsize is stored in this structure.
 */
extern struct mips_cpu mips_cpu;

#undef DEBUG_TLB

/**
 * The icace and dcache size variables are currently set to hard values,
 * as there is no way to size the Lexra caches through software.
 */
__DMEM unsigned long icache_size, dcache_size;
__DMEM unsigned long icache_lsize, dcache_lsize;

extern int r3k_have_wired_reg;  /* in r3k-tlb.c */ 

void tc3162_clear_page(void *);
static void tc3162_copy_page(void *, void *);
extern char except_vec0_r2300;

static inline void flush_dcache_line(unsigned long addr)
{
	__asm__ __volatile__(
		".set noreorder\n\t"
		".set mips3\n\t"
		"cache %1, (%0)\n\t"
		".set mips0\n\t"
		".set reorder"
		:
		: "r" (addr),
		  "i" (Hit_Writeback_Inv_D));
}

/**
 * This function is used to flush a region of the instruction cache. Because
 * Lexra processors do not accomodate instruction cache flushing using 
 * traditional MIPS methods, this function invalidates the entire instruction 
 * cache. This is accomplished using the Lexra specific IINVAL bit of the 
 * coprocessor 0 cache control register.
 * @param start The start of the memory range to flush. Ignored.
 * @param size The size in bytes of the memory range to flush. Ignored.
 */
static void __tc3162_flush_icache_range(unsigned long start,
				    unsigned long end)
{
	unsigned int controlReg;
	unsigned long flags;

	save_and_cli(flags);

	controlReg = read_32bit_cp0_register(CP0_CCTL);

	write_32bit_cp0_register(CP0_CCTL, (controlReg & ~CCTL_IINVAL));
	write_32bit_cp0_register(CP0_CCTL, (controlReg | CCTL_IINVAL));

	/* delay to allow cache to be flushed */
	__asm__ __volatile__(".set\tnoreorder\n\t"
			     "nop\n\t"
			     "nop\n\t"
			     "nop\n\t"
			     "nop\n\t"
			     "nop\n\t"
			     "nop\n\t"
			     "nop\n\t"
			     "nop\n\t" "nop\n\t" ".set\treorder\n\t");

	write_32bit_cp0_register(CP0_CCTL, controlReg);
	restore_flags(flags);
}

/**
 * This function is used to flush a region of the data cache. To accomplish this
 * with a Lexra processor, we read from the corresponding uncached space in
 * KSEG1.
 * @param start The start of the memory range to flush.
 * @param size The size in bytes of the memory range to flush.
 */
static void __tc3162_flush_dcache_range(unsigned long start,
				    unsigned long size)
{
	//unsigned long flags;
	register volatile int *ptr;
	register volatile int *end;
	register int dummy;

	/* Don't have to do anything to KSEG1 addresses */
	if (start == (unsigned long) KSEG1ADDR(start))
		return;


	ptr = (volatile int *) KSEG1ADDR(start & 0xfffffff0);
	end = (volatile int *) KSEG1ADDR(start + size);

	//save_and_cli(flags);
	while (ptr < end) {
		dummy = *ptr;
		ptr += 4;
	}
	//restore_flags(flags);
}


/**
 * This function is used to flush the entire instruction cache. This is
 * done by calling tc3162_flush_cache_range() and specifying the whole of KSEG0 as
 * the range.
 */
static inline void tc3162_flush_cache_all(void)
{
	__tc3162_flush_icache_range(KSEG0, KSEG0 + icache_size);
	__tc3162_flush_dcache_range(KSEG0, KSEG0 + icache_size);
}

/*
 * This function is used to flush a portion of the instruction cache specified
 * by a mm_struct. For the Lexra processor we can only flush the entire cache 
 * so this is done by calling tc3162_flush_cache_range() and specifying the whole 
 * of KSEG0 as the range.
 * @param mm A mm_struct pointer.
 */
static void tc3162_flush_cache_mm(struct mm_struct *mm)
{
	if (mm->context != 0) {
		__tc3162_flush_icache_range(KSEG0, KSEG0 + icache_size);
	}
}

/**
 * This function is used to flush a portion of the instruction cache specified
 * by a mm_struct. For the Lexra processor we can only flush the entire cache 
 * so this is done by calling tc3162_flush_cache_range() and specifying the whole 
 * of KSEG0 as the range.
 * @param mm A mm_struct pointer.
 * @param start Start of memory region to flush.
 * @param end End of memory region to flush.
 */
static void tc3162_flush_cache_range(struct mm_struct *mm,
				 unsigned long start, unsigned long end)
{
	struct vm_area_struct *vma;

	if (mm->context == 0)
		return;

	start &= PAGE_MASK;

	vma = find_vma(mm, start);

	if (!vma)
		return;
	/* we are always going to just flush the entire cache for now */

	__tc3162_flush_icache_range(KSEG0, KSEG0 + icache_size);

}

/**
 * This function is used to flush a portion of the instruction cache specified
 * by a vm_area_struct. For the Lexra processor we can only flush the entire 
 * cache so this is done by calling tc3162_flush_cache_range() and specifying the 
 * whole of KSEG0 as the range.
 * @param vma A vm_area_struct pointer.
 * @param page The page to flush.
 */
static void tc3162_flush_cache_page(struct vm_area_struct *vma,
				unsigned long page)
{
	struct mm_struct *mm = vma->vm_mm;

	if (mm->context == 0)
		return;

	__tc3162_flush_icache_range(KSEG0, KSEG0 + icache_size);
}

/**
 * Lexra has a write through cache, therefore, this is an empty
 * function in the Lexra Linux implementation.
 */
static void tc3162_flush_page_to_ram(struct page *page)
{
	/* Nothing to be done */
}


/**
 * This function is used to flush a portion of the instruction cache specified
 * by a vm_area_struct. For the Lexra processor we can only flush the entire 
 * cache so this is done by calling tc3162_flush_icache_range() and specifying the 
 * whole of KSEG0 as the range.
 * @param vma A vm_area_struct pointer.
 * @param page Pointer to a page structure.
 */
static void tc3162_flush_icache_page(struct vm_area_struct *vma,
				 struct page *page)
{
	struct mm_struct *mm = vma->vm_mm;
	unsigned long physpage;

	if (mm->context == 0)
		return;

	if (!(vma->vm_flags & VM_EXEC))
		return;

	physpage = (unsigned long) page_address(page);
	if (physpage)
		__tc3162_flush_icache_range(KSEG0, KSEG0 + icache_size);
}

/**
 * Traditional R3000 processors flush the first three
 * words ot the cache starting at addr. We must flush
 * the entire cache.
 * @param addr Variable specifying the beginning of the  flush.
 */
static void tc3162_flush_cache_sigtramp(unsigned long addr)
{
	/**
	 * Traditional R3000 processors flush the first three
	 * words ot the cache starting at addr. We must flush
	 * the entire cache.
	 */
	__tc3162_flush_icache_range(addr, KSEG0 + icache_size);
}

/**
 * This function is used to flush and invalidate the memory used for DMA 
 * accesses.  Because Lexra uses a write through cache, it is only necessary 
 * to flush the specified range of the data cache.
 * @param start The start of the memory range used in the DMA transfer.
 * @param size The size in bytes of the DMA transfer.
 */
__IMEM static void __tc3162_dma_cache_wback_inv(unsigned long start,
				     unsigned long size)
{
#if 1
	unsigned long end, a;
	unsigned int flags;

	/* Don't have to do anything to KSEG1 addresses */
	if (start == (unsigned long) KSEG1ADDR(start))
		return;

	if (size >= dcache_size) {
		flush_cache_all();
	} else {
		save_and_cli(flags);

		a = start & ~(dcache_lsize - 1);
		end = (start + size) & ~(dcache_lsize - 1);
		while (1) {
			flush_dcache_line(a); /* Hit_Writeback_Inv_D */
			if (a == end) break;
			a += dcache_lsize;
		}

		restore_flags(flags);
	}
#else
	__tc3162_flush_dcache_range(start, size);
#endif
}

static void __init probe_tlb(void)
{
	int i;
	unsigned long temp;

	mips_cpu.tlbsize = 32;
	temp = get_entryhi();
	for (i=63; i>0; i=i-8) {
		set_index(i<<8);
		set_entryhi(0xaaaaaa80);
		tlb_write_indexed();
		tlb_read();
		if (get_entryhi() == 0xaaaaaa80)
		{
		    mips_cpu.tlbsize = (i + 1);
		    break;
		}
	};
	set_entryhi(temp);
	printk("%d entry TLB.\n", mips_cpu.tlbsize);
}

unsigned long __init tc3162_cache_size(unsigned long ca_flags)
{
	unsigned long flags, status, dummy, size;
	volatile unsigned long *p;
	p = (volatile unsigned long *) KSEG0;
	flags = read_32bit_cp0_register(CP0_STATUS);
	/* isolate cache space */
	write_32bit_cp0_register(CP0_STATUS, (ca_flags|flags)&~ST0_IEC);
	*p = 0xa5a55a5a;
	dummy = *p;
	status = read_32bit_cp0_register(CP0_STATUS);

	if (dummy != 0xa5a55a5a || (status & ST0_CM)) {
		size = 0;
	} else {
		for (size = 128; size <= 0x40000; size <<= 1)
			*(p + size) = 0;
		*p = -1;
		for (size = 128;
		     (size <= 0x40000) && (*(p + size) == 0);
		     size <<= 1)
			;
		if (size > 0x40000)
			size = 0;
	}
	write_32bit_cp0_register(CP0_STATUS, flags);
	return size * sizeof(*p);
}

unsigned long __init tc3162_cache_lsize(unsigned long ca_flags)
{
	unsigned long flags, status, lsize, i;
	volatile unsigned long *p;
	p = (volatile unsigned long *) KSEG0;
	flags = read_32bit_cp0_register(CP0_STATUS);
	/* isolate cache space */
	write_32bit_cp0_register(CP0_STATUS, (ca_flags|flags)&~ST0_IEC);

	for (i = 0; i < 128; i++)
		*(p + i) = 0;
	*(volatile unsigned char *)p = 0;
	for (lsize = 1; lsize < 128; lsize <<= 1) {
		*(p + lsize);
		status = read_32bit_cp0_register(CP0_STATUS);
		if (!(status & ST0_CM))
			break;
	}
	for (i = 0; i < 128; i += lsize)
		*(volatile unsigned char *)(p + i) = 0;
	write_32bit_cp0_register(CP0_STATUS, flags);
	return lsize * sizeof(*p);
}
	
static void __init tc3162_probe_cache(void)
{
	dcache_size = 1024*4;
	if (dcache_size)
		dcache_lsize = 16;
	icache_size = 1024*8;
	if (icache_size)
		icache_lsize = 16;
}

/**
 * This function initializes all of the Lexra memory management functions.
 * This involves mapping the kernel cache management function pointers
 * to the Lexra specific functions, as well as setting the icache and
 * dcache size variables.
 */
void __init ld_mmu_tc3162(void)
{
        r3k_have_wired_reg = 1;  

	_clear_page = tc3162_clear_page;
	_copy_page = tc3162_copy_page;

	/*
	 * These variables are unused on this architecture.
	 */
	tc3162_probe_cache();
	
	_flush_cache_all = tc3162_flush_cache_all;
	___flush_cache_all = tc3162_flush_cache_all; 
	_flush_cache_mm = tc3162_flush_cache_mm;
	_flush_cache_range = tc3162_flush_cache_range;
	_flush_cache_page = tc3162_flush_cache_page;
	_flush_cache_sigtramp = tc3162_flush_cache_sigtramp;
	_flush_page_to_ram = tc3162_flush_page_to_ram;
	_flush_icache_page = tc3162_flush_icache_page;

	_flush_icache_range = __tc3162_flush_icache_range;
//	_flush_dcache_range = __tc3162_flush_dcache_range;

	_dma_cache_wback_inv = __tc3162_dma_cache_wback_inv;

	probe_tlb();

	_flush_cache_all();
	flush_tlb_all();

	memcpy((void *)KSEG0, &except_vec0_r2300, 0x80);
	flush_icache_range(KSEG0, KSEG0 + 0x80);
	printk("Primary instruction cache %dkb, linesize %d bytes\n",
			(int) (icache_size >> 10), (int) icache_lsize);
	printk("Primary data cache %dkb, linesize %d bytes\n",
			(int) (dcache_size >> 10), (int) dcache_lsize);		 
}

/* From r2300.c */

/**
 * Assebly language function to clear a page of memory.
 * @param page Pointer to the page.
 */
void tc3162_clear_page(void *page)
{
	__asm__ __volatile__(".set\tnoreorder\n\t"
			     ".set\tnoat\n\t"
			     "addiu\t$1,%0,%2\n"
			     "1:\tsw\t$0,(%0)\n\t"
			     "sw\t$0,4(%0)\n\t"
			     "sw\t$0,8(%0)\n\t"
			     "sw\t$0,12(%0)\n\t"
			     "addiu\t%0,32\n\t"
			     "sw\t$0,-16(%0)\n\t"
			     "sw\t$0,-12(%0)\n\t"
			     "sw\t$0,-8(%0)\n\t"
			     "bne\t$1,%0,1b\n\t"
			     "sw\t$0,-4(%0)\n\t"
			     ".set\tat\n\t"
			     ".set\treorder":"=r"(page):"0"(page),
			     "I"(PAGE_SIZE):"$1", "memory");
}


/**
 * This function is used to copy a page (4096 bytes) from the location
 * pointed to by 'from' to the location pointed to by 'to'.
 * @param to Location to copy to.
 * @param from Location to copy from.
 */
static void tc3162_copy_page(void *to, void *from)
{
	unsigned long dummy1, dummy2;
	unsigned long reg1, reg2, reg3, reg4;

	/* This function copies a 4 kb page from addr 'from' to the addr
	 * pointed to by 'to'.
	 */
	__asm__ __volatile__(".set\tnoreorder\n\t"
			     ".set\tnoat\n\t"
			     "addiu\t$1,%0,%8\n"
			     "1:\tlw\t%2,(%1)\n\t"
			     "lw\t%3,4(%1)\n\t"
			     "lw\t%4,8(%1)\n\t"
			     "lw\t%5,12(%1)\n\t"
			     "sw\t%2,(%0)\n\t"
			     "sw\t%3,4(%0)\n\t"
			     "sw\t%4,8(%0)\n\t"
			     "sw\t%5,12(%0)\n\t"
			     "lw\t%2,16(%1)\n\t"
			     "lw\t%3,20(%1)\n\t"
			     "lw\t%4,24(%1)\n\t"
			     "lw\t%5,28(%1)\n\t"
			     "sw\t%2,16(%0)\n\t"
			     "sw\t%3,20(%0)\n\t"
			     "sw\t%4,24(%0)\n\t"
			     "sw\t%5,28(%0)\n\t"
			     "addiu\t%0,64\n\t"
			     "addiu\t%1,64\n\t"
			     "lw\t%2,-32(%1)\n\t"
			     "lw\t%3,-28(%1)\n\t"
			     "lw\t%4,-24(%1)\n\t"
			     "lw\t%5,-20(%1)\n\t"
			     "sw\t%2,-32(%0)\n\t"
			     "sw\t%3,-28(%0)\n\t"
			     "sw\t%4,-24(%0)\n\t"
			     "sw\t%5,-20(%0)\n\t"
			     "lw\t%2,-16(%1)\n\t"
			     "lw\t%3,-12(%1)\n\t"
			     "lw\t%4,-8(%1)\n\t"
			     "lw\t%5,-4(%1)\n\t"
			     "sw\t%2,-16(%0)\n\t"
			     "sw\t%3,-12(%0)\n\t"
			     "sw\t%4,-8(%0)\n\t"
			     "bne\t$1,%0,1b\n\t"
			     "sw\t%5,-4(%0)\n\t"
			     ".set\tat\n\t"
			     ".set\treorder":"=r"(dummy1), "=r"(dummy2),
			     "=&r"(reg1), "=&r"(reg2), "=&r"(reg3),
			     "=&r"(reg4):"0"(to), "1"(from),
			     "I"(PAGE_SIZE));
}
